#include <PID_v1_bc.h>
#include <Wire.h>
#include <Adafruit_MotorShield.h>
/**
 * Arduino-Steuerprogramm für Fischertechnik-Parallel-Tripteron
 * 
 * (c) Florian Bauer  - September 2024
 * 
 * Hardware:
 * - 3 XM-Motoren mit Seilantrieb
 *    - Vordere Achse   Mot1
 *    - links hinten    Mot3
 *    - rechts hinten   Mot2
 *    
 * - Arduino Uno
 * - Motor Shield v2.0
 * - Sharp Reflex-Sensoren
 *   - für hintere Achsen: GP 2Y0A41 SK0F 2 - 55cm    -> A1, A2
 *   - für vordere Achse: GP2Y0A21YK0F 10-80cm        -> A0
 *   
 */
 
// Create the motor shield object with the default I2C address
Adafruit_MotorShield AFMS = Adafruit_MotorShield();

// Select which 'port' M1, M2, M3 or M4. In this case, M1
Adafruit_DCMotor *mot1 = AFMS.getMotor(1);
// Select which 'port' M1, M2, M3 or M4. In this case, M1
Adafruit_DCMotor *mot2 = AFMS.getMotor(2);

Adafruit_DCMotor *mot3 = AFMS.getMotor(3);

#include <Wire.h>
#include <VL53L0X.h>
VL53L0X sensor;

#define PID_P 12

#define WINDOW_SIZE 4
float buf2[WINDOW_SIZE];
int buf2_i=0;

float avg(float y, float * buf,int w, int j)
{
  int jp=j;
  float v=0.0;
  for(int i=0;i<w;i++)
  {
    v += buf[jp];
    jp++;
    jp=jp% w;
  }
  return v / (float) w;
}

String readString; //This while store the user input data


double input1Old=0, input1 = 0, output1 = 0, setpoint1 = 0, input1_0 = 0;
double input2Old=0, input2 = 0, output2 = 0, setpoint2 = 0,input2_0 = 0;
double input3Old=0, input3 = 0, output3 = 0, setpoint3 = 0,input3_0 = 0;
int limit = 222;


volatile float p1,p2,p3;
int  distSensR [] = {625,612,530,451,387,341,307,272,250,228,210,195,182,174,162,155,142,135 };
int  distSensL [] = {584,617,520,440,380,332,294,262,238,220,203,186,175,163,155,145,132,132 };
int  distSensM [] = {533,533,533,533,643,644,640,580,520,473,430,395,367,340,320,320,320,320 };
//int  distSensR [] = {650,616,568,523,481,440,407,377,351,326,307,291,276,260,247,233,222,206,194,184,174}; // 0= 6cm 80cm sensor


float readDistSensor()
{
  int val;

    buf2_i ++;
    buf2_i = buf2_i % WINDOW_SIZE;
    val=sensor.readRangeSingleMillimeters();
    buf2[buf2_i] = (float)val;
//    Serial.print("Distance in mm: ");
 //   Serial.print(val); Serial.print(" "); Serial.println(avg( (float) val,buf,WINDOW_SIZE,buf_i));
  
  return avg( (float) val,buf2,WINDOW_SIZE,buf2_i);
}
 float posToDistSens( float x, int * distSens,int n)
{
   float m,y,ya,yb;
   int xi = (int) (x );

   if (xi>= n-1) return -1.0;
   ya = distSens[xi];
   yb = distSens[xi+1];
   m = (yb-ya);
   y = (x - (float)xi) * m + ya;
  /* Serial.print("x-xi=");
   Serial.print(x - (float) xi); Serial.print(" m=");
   Serial.print(m);   Serial.print(" y=");
   Serial.print(y); Serial.println(" ");
  */
   return y;
}

float distSensToPos(float s, int * distSens, int n)
{
  int j;
  for(j=0;j<n-1;j++)
  {
    if(s <= distSens[j] && s >= distSens[j+1])
    {
      return (s - distSens[j]) / (distSens[j+1]-distSens[j]) + j;
    }
  }
  return -1.0;
}
void setup() {
  // put your setup code here, to run once:
  Serial.begin(250000); //initialize serial comunication 
  Serial.println("Setup Parallel Tripteron");
  
 
  AFMS.begin();

  for(int i=0;i<WINDOW_SIZE;i++) buf2[i]=0.0;
  buf2_i=0;
  Serial.println("VL53L0X API Continuous\n");

  // start time of flight distance sensor
  //sensor.init();
  // sensor.setTimeout(500);
  
  int val1 = analogRead(A0);  // read the input pin
  int val2 = analogRead(A1);  // read the input pin
  //int val2 = readDistSensor();
  
   int val3 = analogRead(A2);  // read the input pin
  p1=setpoint1=distSensToPos(val1,distSensM,21)*10;
  p2=setpoint2=distSensToPos(val2,distSensR,21)*10;;
   p3=setpoint3=distSensToPos(val3,distSensL,21)*10;


}

void pwmOut(int i,int out) {     
                
  if (out > 0) {                         // if REV > encoderValue motor move in forward direction.    
     switch(i)
     {
      case 0: mot1->setSpeed(out); mot1->run(FORWARD);  break;
       case 1:  mot2->setSpeed(out); mot2->run(FORWARD); break;
        case 2: mot3->setSpeed(out); mot3->run(FORWARD); break;
     }
  }
  else {
     switch(i) {
     case 0 : mot1->setSpeed(abs(out)); mot1->run(BACKWARD); break;
      case 1: mot2->setSpeed(abs(out)); mot2->run(BACKWARD);  break;
       case 2: mot3->setSpeed(abs(out)); mot3->run(BACKWARD);  break;
     }

  }
}

int analogReadAverage(int ch, int n)
{
  long val=0;
  for(int i=0;i<n;i++)
  {
    val += analogRead(ch);
  }
  return val / n;
}
int runmode=0; // initially all Motors stopped!





int clipLimits(int axis,int val)
{
 switch(axis)
 {
  case 1: // Middle
  if(val<70) return 70;
  if(val > 170) return 170;
  break;
  case 2: //Right
  if(val<20) return 20;
  if(val >130) return 130;
  break;
  case 3: // Left
  if(val<20) return 20;
  if(val > 130) return 130;
  break;
 }
 return val;
}

void showHelp()
{
        Serial.println("Parallel Tripteron Commands");
        Serial.println("r      : Initialposition anfahren");
        Serial.println("0      : Setze aktuelle Position als Soll-Position");
        Serial.println("s      : Motoren ein / aus");
        Serial.println("m <m>,<r>,<l> : Stelle Achsen auf position <m>,<r>,<l>");
        Serial.println("1 <m>  : Fahre Motor 1 (mittlere Achse) auf Position <m>");
        Serial.println("2 <r>  : Fahre Motor 2 (rechte Achse) auf Position <r>");
        Serial.println("3 <v>  : Fahre Motor 3 (linke Achse) auf Position <l>");
        Serial.println("x <v>  : Verfahre um <v> x 10 nach rechts");
        Serial.println("X <v>  : Verfahre um <v> x 10 nach links");
        Serial.println("y <v>  : Verfahre um <v> x 10 rückwärts");
        Serial.println("Y <v>  : Verfahre um <v> x 10 vorwärts");
        Serial.println("z <v>  : Verfahre um <v> x 10 nach unten");
        Serial.println("Z <v>  : Verfahre um <v> x 10 nach oben");       
        delay(4000);
        }
        
void loop() {
  char buf[100];
  char * ptr;
  
  // put your main code here, to run repeatedly:
  int val1 = analogReadAverage(A0,16);  // read the input pin
  int val2 = analogReadAverage(A1,16);  // read the input pin
  
  //int val2 = readDistSensor();
  int val3 = analogReadAverage(A2,16);  // read the input pin
  input1 = 10 * distSensToPos(val1,distSensM,21) ;
  input2 = 10 * distSensToPos(val2,distSensR,21) ;
  input3 = 10 * distSensToPos(val3,distSensL,21) ;
  
  while (Serial.available()) { //Check if the serial data is available.
    delay(3);                  // a small delay
    char c = Serial.read();  // storing input data
    readString += c;         // accumulate each of the characters in readString
  }
  //input1 = val1;
  //input2 = val2;
  //input3 = val3;
   if (readString.length() >0) { //Verify that the variable contains information  
      readString.toCharArray(buf, 100);
      if (strncmp(buf,"s",1)==0)
      {
        runmode=1-runmode;
      }
       if (strncmp(buf,"?",1)==0 || strncmp(buf,"h",1)==0 )
      {  
        showHelp();  
      }
       if (strncmp(buf,"r",1)==0)
      {  
        p1=setpoint1=100;
        p2=setpoint2=40;
        p3=setpoint3=40;
      }
      if (strncmp(buf,"0",1)==0)
      {  
        p1=setpoint1= distSensToPos(analogRead(A0),distSensM,21);
        p2=setpoint2=distSensToPos(analogRead(A1),distSensR,21);;
        //p2=setpoint2=readDistSensor(); 
        p3=setpoint3=distSensToPos(analogRead(A2),distSensL,21);;
      }
        if (strncmp(buf,"p",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        float p=atof(ptr);
        Serial.println( posToDistSens(p,distSensR,18));
        
      }
      if (strncmp(buf,"d",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        float d=atof(ptr);
        Serial.println( distSensToPos(d,distSensR,18));
        
      }
       if (strncmp(buf,"1",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        p1=atof(ptr);setpoint1=p1;
      }
       if (strncmp(buf,"2",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        p2=atof(ptr);setpoint2=p2;
      }
             if (strncmp(buf,"3",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        p3=atof(ptr);setpoint3=p3;
      }
      if (strncmp(buf,"x",1)==0)
      {  
        ptr = strtok(buf+1, ",");
        float n=atof(ptr);
        
        p2 = p2 + n;
        p3 = p3 - n;
         setpoint2=p2;
            setpoint3=p3;
      }
      if (strncmp(buf,"X",1)==0)
      {  
         ptr = strtok(buf+1, ",");
        float n=atof(ptr);
        p2 = p2 - n;
        p3 = p3 + n;
         setpoint2=p2;
            setpoint3=p3;
      }
      if (strncmp(buf,"y",1)==0)
      {  
         ptr = strtok(buf+1, ",");
        float n=atof(ptr);
        p1 = p1 - n;
        p2 = p2 - n;
        p3 = p3 - n;
        setpoint1=p1;
         setpoint2=p2;
            setpoint3=p3;
      }
      if (strncmp(buf,"Y",1)==0)
      {  
          ptr = strtok(buf+1, ",");
        float n=atof(ptr);
         p1 = p1 + n;
        p2 = p2 + n;
        p3 = p3 + n;
        setpoint1=p1;
         setpoint2=p2;
            setpoint3=p3;
      }
        if (strncmp(buf,"Z",1)==0)
      {  
                  ptr = strtok(buf+1, ",");
        int n=atof(ptr);
        p1 = p1 + n;
        p2 = p2 - n;
        p3 = p3 - n;
        setpoint1=p1;
        setpoint2=p2;
        setpoint3=p3;
      }
      if (strncmp(buf,"z",1)==0) 
      {  
        ptr = strtok(buf+1, ",");
        float n=atof(ptr);
         p1 = p1 - n;
        p2 = p2 + n;
        p3 = p3 + n;
        setpoint1=p1;
        setpoint2=p2;
        setpoint3=p3;
      }
      if (strncmp(buf,"m",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        p1=atof(ptr);
     // Serial.println(ptr);
        if (ptr != NULL) {
          ptr = strtok(NULL,",");
          if( ptr != NULL)
          {
            p2=atof(ptr);
             if (ptr != NULL) {
          ptr = strtok(NULL,",");
          if( ptr != NULL)
          {
            p3=atof(ptr);
            Serial.print(p1); Serial.print(" "); 
            Serial.print(p2); Serial.print(" ");
            Serial.print(p3); Serial.print(" ");
            Serial.println();
            setpoint1=p1;
            setpoint2=p2;
            setpoint3=p3;
          }
             } } } } 
 if (strncmp(buf,"M",1)==0)
      {  
        
        ptr = strtok(buf+1, ",");
        float q1=atof(ptr);
     // Serial.println(ptr);
        if (ptr != NULL) {
          ptr = strtok(NULL,",");
          if( ptr != NULL)
          {
            float q2=atof(ptr);
             if (ptr != NULL) {
          ptr = strtok(NULL,",");
          if( ptr != NULL)
          {
            float q3=atof(ptr);
            p1 = posToDistSens(q1,distSensM,18);
            p2 = posToDistSens(q2,distSensR,18);
            p3 = posToDistSens(q3,distSensL,18);
            
            Serial.print(p1); Serial.print(" "); 
            Serial.print(p2); Serial.print(" ");
            Serial.print(p3); Serial.print(" ");
            Serial.println();
            setpoint1=p1;
            setpoint2=p2;
            setpoint3=p3;
          }
             } } } } 
          //   delay(1000);
              readString="";
  }
  p1= clipLimits(1,p1); 
  setpoint1=p1;
  p2= clipLimits(2,p2); 
 setpoint2=p2;
 p3= clipLimits(3,p3); 
 setpoint3=p3;

  

/*
  int delta = setpoint1-input1;
  if(delta>3)
  {
   
     if(delta>150) delta=150;
     if(delta<50) delta=60;
     output1= - delta;
  }
  else if (delta< -3)
  {
  
    if(delta < -150) delta= -150;
    if(delta > -50) delta= -60;
    output1 =  -delta;
  }
  */
     output1 = PID_P*(input1 - setpoint1);
   if (output1 > limit) output1=limit;
   if (output1 < -limit) output1 = -limit;
  
  if(runmode)
    pwmOut(0,output1);
  else
    pwmOut(0,0);


  /*
 int delta2 = setpoint2-input2;
  if(delta2>3)
  {
    
     if(delta2>150) delta2=150;
     if(delta2<50) delta2=60;
     output2= - delta2;
  }
  else if (delta2< -3)
  {
   
    if(delta2 < -150) delta2= -150;
    if(delta2 > -50) delta2= -60;
    output2 =  -delta2;
  }
  */

   output2 = PID_P*(input2 - setpoint2);
   if (output2 > limit) output2=limit;
   if (output2 < -limit) output2 = -limit;
  
  if(runmode)
    pwmOut(1,output2);
  else
    pwmOut(1,0);

    /*
  int delta3 = setpoint3-input3;
  if(delta3>3)
  {
    
     if(delta3>150) delta3=150;
     if(delta3<50) delta3=60;
     output3= - delta3;
  }
  else if (delta3< -3)
  {
    
    if(delta3 < -150) delta3= -150;
    if(delta3 > -50) delta3= -60;
    output3 =  -delta3;
  }
 */
   
   output3 = 8*(input3 - setpoint3);
   if (output3 > limit) output3=limit;
   if (output3 < -limit) output3 = -limit;
  if(runmode)
    pwmOut(2,output3);
  else
    pwmOut(2,0);
    
 
 // pwmOut(1,output2);
 // Serial.println(output);
  
  if(runmode==0)
  Serial.print("HALT ");
  else
  Serial.print("RUN  ");
   Serial.print(setpoint1);      Serial.print(" ");
  Serial.print(input1);          // debug value
  Serial.print(" ");
    //Serial.print(output1);          // debug value
  Serial.print(" ");
   Serial.print(setpoint2);      Serial.print(" ");
 // Serial.print(val2);          // debug value
  Serial.print(" ");
   Serial.print(input2);          // debug value
  Serial.print(" ");
//  Serial.print(output2);          // debug value
 // Serial.print(" ");
 //  Serial.print(p3);      Serial.print(" ");
   Serial.print(setpoint3);      Serial.print(" ");
  Serial.print(input3);          // debug value
  Serial.print(" ");
  //  Serial.print(output3);          // debug value
  Serial.println(" ");
    
}
